import sys
from contextlib import ExitStack, contextmanager
from enum import Enum, auto
from types import SimpleNamespace, FrameType, ModuleType
from typing import List
from bdb import BdbQuit
from inspect import currentframe

import cadquery as cq
from PyQt5 import QtCore
from PyQt5.QtCore import (
    Qt,
    QObject,
    pyqtSlot,
    pyqtSignal,
    QEventLoop,
    QAbstractTableModel,
)
from PyQt5.QtWidgets import QAction, QTableView

from logbook import info
from path import Path
from pyqtgraph.parametertree import Parameter
from ..icons import icon
from random import randrange as rrr, seed

from ..cq_utils import find_cq_objects, reload_cq
from ..mixins import ComponentMixin

DUMMY_FILE = "<cq_editor-string>"


class DbgState(Enum):

    STEP = auto()
    CONT = auto()
    STEP_IN = auto()
    RETURN = auto()


class DbgEevent(object):

    LINE = "line"
    CALL = "call"
    RETURN = "return"


class LocalsModel(QAbstractTableModel):

    HEADER = ("Name", "Type", "Value")

    def __init__(self, parent):

        super(LocalsModel, self).__init__(parent)
        self.frame = None

    def update_frame(self, frame):

        self.frame = [
            (k, type(v).__name__, str(v))
            for k, v in frame.items()
            if not k.startswith("_")
        ]

    def rowCount(self, parent=QtCore.QModelIndex()):

        if self.frame:
            return len(self.frame)
        else:
            return 0

    def columnCount(self, parent=QtCore.QModelIndex()):

        return 3

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if role == Qt.DisplayRole and orientation == Qt.Horizontal:
            return self.HEADER[section]
        return QAbstractTableModel.headerData(self, section, orientation, role)

    def data(self, index, role):
        if role == QtCore.Qt.DisplayRole:
            i = index.row()
            j = index.column()
            return self.frame[i][j]
        else:
            return QtCore.QVariant()


class LocalsView(QTableView, ComponentMixin):

    name = "Variables"

    def __init__(self, parent):

        super(LocalsView, self).__init__(parent)
        ComponentMixin.__init__(self)

        header = self.horizontalHeader()
        header.setStretchLastSection(True)

        vheader = self.verticalHeader()
        vheader.setVisible(False)

    @pyqtSlot(dict)
    def update_frame(self, frame):

        model = LocalsModel(self)
        model.update_frame(frame)

        self.setModel(model)


class Debugger(QObject, ComponentMixin):

    name = "Debugger"

    preferences = Parameter.create(
        name="Preferences",
        children=[
            {"name": "Reload CQ", "type": "bool", "value": False},
            {"name": "Add script dir to path", "type": "bool", "value": True},
            {"name": "Change working dir to script dir", "type": "bool", "value": True},
            {"name": "Reload imported modules", "type": "bool", "value": True},
        ],
    )

    sigRendered = pyqtSignal(dict)
    sigLocals = pyqtSignal(dict)
    sigTraceback = pyqtSignal(object, str)

    sigFrameChanged = pyqtSignal(object)
    sigLineChanged = pyqtSignal(int)
    sigLocalsChanged = pyqtSignal(dict)
    sigCQChanged = pyqtSignal(dict, bool)
    sigDebugging = pyqtSignal(bool)

    _frames: List[FrameType]
    _stop_debugging: bool

    def __init__(self, parent):

        super(Debugger, self).__init__(parent)
        ComponentMixin.__init__(self)

        self.inner_event_loop = QEventLoop(self)

        self._actions = {
            "Run": [
                QAction(
                    icon("run"), "Render", self, shortcut="F5", triggered=self.render
                ),
                QAction(
                    icon("debug"),
                    "Debug",
                    self,
                    checkable=True,
                    shortcut="ctrl+F5",
                    triggered=self.debug,
                ),
                QAction(
                    icon("arrow-step-over"),
                    "Step",
                    self,
                    shortcut="ctrl+F10",
                    triggered=lambda: self.debug_cmd(DbgState.STEP),
                ),
                QAction(
                    icon("arrow-step-in"),
                    "Step in",
                    self,
                    shortcut="ctrl+F11",
                    triggered=lambda: self.debug_cmd(DbgState.STEP_IN),
                ),
                QAction(
                    icon("arrow-continue"),
                    "Continue",
                    self,
                    shortcut="ctrl+F12",
                    triggered=lambda: self.debug_cmd(DbgState.CONT),
                ),
            ]
        }

        self._frames = []
        self._stop_debugging = False

    def get_current_script(self):

        return self.parent().components["editor"].get_text_with_eol()

    def get_current_script_path(self):

        filename = self.parent().components["editor"].filename
        if filename:
            return Path(filename).absolute()

    def get_breakpoints(self):

        return self.parent().components["editor"].debugger.get_breakpoints()

    def set_breakpoints(self, breakpoints):
        return self.parent().components["editor"].debugger.set_breakpoints(breakpoints)

    def compile_code(self, cq_script, cq_script_path=None):

        try:
            module = ModuleType("__cq_main__")
            if cq_script_path:
                module.__dict__["__file__"] = cq_script_path
            cq_code = compile(cq_script, DUMMY_FILE, "exec")
            return cq_code, module
        except Exception:
            self.sigTraceback.emit(sys.exc_info(), cq_script)
            return None, None

    def _exec(self, code, locals_dict, globals_dict):

        with ExitStack() as stack:
            p = (self.get_current_script_path() or Path("")).absolute().dirname()

            if self.preferences["Add script dir to path"] and p.exists():
                sys.path.insert(0, p)
                stack.callback(sys.path.remove, p)
            if self.preferences["Change working dir to script dir"] and p.exists():
                stack.enter_context(p)
            if self.preferences["Reload imported modules"]:
                stack.enter_context(module_manager())

            exec(code, locals_dict, globals_dict)

    @staticmethod
    def _rand_color(alpha=0.0, cfloat=False):
        # helper function to generate a random color dict
        # for CQ-editor's show_object function
        lower = 10
        upper = 100  # not too high to keep color brightness in check
        if cfloat:  # for two output types depending on need
            return (
                (rrr(lower, upper) / 255),
                (rrr(lower, upper) / 255),
                (rrr(lower, upper) / 255),
                alpha,
            )
        return {
            "alpha": alpha,
            "color": (
                rrr(lower, upper),
                rrr(lower, upper),
                rrr(lower, upper),
            ),
        }

    def _inject_locals(self, module):

        cq_objects = {}

        def _show_object(obj, name=None, options={}):

            if name:
                cq_objects.update({name: SimpleNamespace(shape=obj, options=options)})
            else:
                # get locals of the enclosing scope
                d = currentframe().f_back.f_locals

                # try to find the name
                try:
                    name = list(d.keys())[list(d.values()).index(obj)]
                except ValueError:
                    # use id if not found
                    name = str(id(obj))

                cq_objects.update({name: SimpleNamespace(shape=obj, options=options)})

        def _debug(obj, name=None):

            _show_object(obj, name, options=dict(color="red", alpha=0.2))

        module.__dict__["show_object"] = _show_object
        module.__dict__["debug"] = _debug
        module.__dict__["rand_color"] = self._rand_color
        module.__dict__["log"] = lambda x: info(str(x))
        module.__dict__["cq"] = cq

        return cq_objects, set(module.__dict__) - {"cq"}

    def _cleanup_locals(self, module, injected_names):

        for name in injected_names:
            module.__dict__.pop(name)

    @pyqtSlot(bool)
    def render(self):

        seed(59798267586177)
        if self.preferences["Reload CQ"]:
            reload_cq()

        cq_script = self.get_current_script()
        cq_script_path = self.get_current_script_path()
        cq_code, module = self.compile_code(cq_script, cq_script_path)

        if cq_code is None:
            return

        cq_objects, injected_names = self._inject_locals(module)

        try:
            self._exec(cq_code, module.__dict__, module.__dict__)

            # remove the special methods
            self._cleanup_locals(module, injected_names)

            # collect all CQ objects if no explicit show_object was called
            if len(cq_objects) == 0:
                cq_objects = find_cq_objects(module.__dict__)
            self.sigRendered.emit(cq_objects)
            self.sigTraceback.emit(None, cq_script)
            self.sigLocals.emit(module.__dict__)
        except Exception:
            exc_info = sys.exc_info()
            sys.last_traceback = exc_info[-1]
            self.sigTraceback.emit(exc_info, cq_script)

    @property
    def breakpoints(self):
        return [el[0] for el in self.get_breakpoints()]

    @pyqtSlot(bool)
    def debug(self, value):

        # used to stop the debugging session early
        self._stop_debugging = False

        if value:
            self.previous_trace = previous_trace = sys.gettrace()

            self.sigDebugging.emit(True)
            self.state = DbgState.STEP

            self.script = self.get_current_script()
            cq_script_path = self.get_current_script_path()
            code, module = self.compile_code(self.script, cq_script_path)

            if code is None:
                self.sigDebugging.emit(False)
                self._actions["Run"][1].setChecked(False)
                return

            cq_objects, injected_names = self._inject_locals(module)

            # clear possible traceback
            self.sigTraceback.emit(None, self.script)

            try:
                sys.settrace(self.trace_callback)
                exec(code, module.__dict__, module.__dict__)
            except BdbQuit:
                pass
            except Exception:
                exc_info = sys.exc_info()
                sys.last_traceback = exc_info[-1]
                self.sigTraceback.emit(exc_info, self.script)
            finally:
                sys.settrace(previous_trace)
                self.sigDebugging.emit(False)
                self._actions["Run"][1].setChecked(False)

                if len(cq_objects) == 0:
                    cq_objects = find_cq_objects(module.__dict__)
                self.sigRendered.emit(cq_objects)

                self._cleanup_locals(module, injected_names)
                self.sigLocals.emit(module.__dict__)

                self._frames = []
                self.inner_event_loop.exit(0)
        else:
            self._stop_debugging = True
            self.inner_event_loop.exit(0)

    def debug_cmd(self, state=DbgState.STEP):

        self.state = state
        self.inner_event_loop.exit(0)

    def trace_callback(self, frame, event, arg):

        filename = frame.f_code.co_filename

        if filename == DUMMY_FILE:
            if not self._frames:
                self._frames.append(frame)
            self.trace_local(frame, event, arg)
            return self.trace_callback

        else:
            return None

    def trace_local(self, frame, event, arg):

        lineno = frame.f_lineno

        if event in (DbgEevent.LINE,):
            if (
                self.state in (DbgState.STEP, DbgState.STEP_IN)
                and frame is self._frames[-1]
            ) or (lineno in self.get_breakpoints()):

                if lineno in self.get_breakpoints():
                    self._frames.append(frame)

                self.sigLineChanged.emit(lineno)
                self.sigFrameChanged.emit(frame)
                self.sigLocalsChanged.emit(frame.f_locals)
                self.sigCQChanged.emit(find_cq_objects(frame.f_locals), True)

                self.inner_event_loop.exec_()

        elif event in (DbgEevent.RETURN):
            self.sigLocalsChanged.emit(frame.f_locals)
            self._frames.pop()

        elif event == DbgEevent.CALL:
            func_filename = frame.f_code.co_filename
            if self.state == DbgState.STEP_IN and func_filename == DUMMY_FILE:
                self.sigLineChanged.emit(lineno)
                self.sigFrameChanged.emit(frame)
                self.state = DbgState.STEP
                self._frames.append(frame)

        if self._stop_debugging:
            raise BdbQuit  # stop debugging if requested


@contextmanager
def module_manager():
    """unloads any modules loaded while the context manager is active"""
    loaded_modules = set(sys.modules.keys())

    try:
        yield
    finally:
        new_modules = set(sys.modules.keys()) - loaded_modules
        for module_name in new_modules:
            del sys.modules[module_name]



from sys import platform


from PyQt5.QtWidgets import QWidget, QApplication
from PyQt5.QtCore import pyqtSlot, pyqtSignal, Qt, QPoint

import OCP

from OCP.Aspect import Aspect_DisplayConnection, Aspect_TypeOfTriedronPosition
from OCP.OpenGl import OpenGl_GraphicDriver
from OCP.V3d import V3d_Viewer
from OCP.gp import gp_Trsf, gp_Ax1, gp_Dir
from OCP.AIS import AIS_InteractiveContext, AIS_DisplayMode
from OCP.Quantity import Quantity_Color


ZOOM_STEP = 0.9


class OCCTWidget(QWidget):

    sigObjectSelected = pyqtSignal(list)

    def __init__(self, parent=None):

        super(OCCTWidget, self).__init__(parent)

        self.setAttribute(Qt.WA_NativeWindow)
        self.setAttribute(Qt.WA_PaintOnScreen)
        self.setAttribute(Qt.WA_NoSystemBackground)

        self._initialized = False
        self._needs_update = False
        self._previous_pos = QPoint(
            0, 0  # Keeps track of where the previous mouse position
        )
        self._rotate_step = (
            0.008  # Controls the speed of rotation with the turntable orbit method
        )

        # Orbit method settings
        self._orbit_method = "Turntable"

        # OCCT secific things
        self.display_connection = Aspect_DisplayConnection()
        self.graphics_driver = OpenGl_GraphicDriver(self.display_connection)

        self.viewer = V3d_Viewer(self.graphics_driver)
        self.view = self.viewer.CreateView()
        self.context = AIS_InteractiveContext(self.viewer)

        # Trihedorn, lights, etc
        self.prepare_display()

    def prepare_display(self):

        view = self.view

        params = view.ChangeRenderingParams()
        params.NbMsaaSamples = 8
        params.IsAntialiasingEnabled = True

        view.TriedronDisplay(
            Aspect_TypeOfTriedronPosition.Aspect_TOTP_RIGHT_LOWER, Quantity_Color(), 0.1
        )

        viewer = self.viewer

        viewer.SetDefaultLights()
        viewer.SetLightOn()

        ctx = self.context

        ctx.SetDisplayMode(AIS_DisplayMode.AIS_Shaded, True)
        ctx.DefaultDrawer().SetFaceBoundaryDraw(True)

    def set_orbit_method(self, method):
        """
        Set the orbit method for the OCCT view.
        """

        # Keep track of which orbit method is used
        if method == "Turntable":
            self._orbit_method = "Turntable"
            self.view.SetUp(0, 0, 1)
        elif method == "Trackball":
            self._orbit_method = "Trackball"
        else:
            raise ValueError(f"Unknown orbit method: {method}")

    def wheelEvent(self, event):

        delta = event.angleDelta().y()
        factor = ZOOM_STEP if delta < 0 else 1 / ZOOM_STEP

        self.view.SetZoom(factor)

    def mousePressEvent(self, event):

        pos = event.pos()

        if event.button() == Qt.LeftButton:
            # Used to prevent drag selection of objects
            self.pending_select = True
            self.left_press = pos

            # We only start the rotation if the orbit method is set to Trackball
            if self._orbit_method == "Trackball":
                self.view.StartRotation(pos.x(), pos.y())
        elif event.button() == Qt.RightButton:
            self.view.StartZoomAtPoint(pos.x(), pos.y())

        self._previous_pos = pos

    def mouseMoveEvent(self, event):

        pos = event.pos()
        x, y = pos.x(), pos.y()

        # Check for mouse drag rotation
        if event.buttons() == Qt.LeftButton:
            # Set the rotation differently based on the orbit method
            if self._orbit_method == "Trackball":
                self.view.Rotation(x, y)
            elif self._orbit_method == "Turntable":
                # Control the turntable rotation manually
                delta_x, delta_y = (
                    x - self._previous_pos.x(),
                    y - self._previous_pos.y(),
                )
                cam = self.view.Camera()
                z_rotation = gp_Trsf()
                z_rotation.SetRotation(
                    gp_Ax1(cam.Center(), gp_Dir(0, 0, 1)), -delta_x * self._rotate_step
                )
                cam.Transform(z_rotation)
                self.view.Rotate(0, -delta_y * self._rotate_step, 0)

            # If the user moves the mouse at all, the selection will not happen
            if abs(x - self.left_press.x()) > 2 or abs(y - self.left_press.y()) > 2:
                self.pending_select = False

        elif event.buttons() == Qt.MiddleButton:
            self.view.Pan(
                x - self._previous_pos.x(), self._previous_pos.y() - y, theToStart=True
            )

        elif event.buttons() == Qt.RightButton:
            self.view.ZoomAtPoint(self._previous_pos.x(), y, x, self._previous_pos.y())

        self._previous_pos = pos

    def mouseReleaseEvent(self, event):

        if event.button() == Qt.LeftButton:
            pos = event.pos()
            x, y = pos.x(), pos.y()

            # Only make the selection if the user has not moved the mouse
            if self.pending_select:
                self.context.MoveTo(x, y, self.view, True)
                self._handle_selection()

    def _handle_selection(self):

        self.context.Select(True)
        self.context.InitSelected()

        selected = []
        if self.context.HasSelectedShape():
            selected.append(self.context.SelectedShape())

        self.sigObjectSelected.emit(selected)

    def paintEngine(self):

        return None

    def paintEvent(self, event):

        if not self._initialized:
            self._initialize()
        else:
            self.view.Redraw()

    def showEvent(self, event):

        super(OCCTWidget, self).showEvent(event)

    def resizeEvent(self, event):

        super(OCCTWidget, self).resizeEvent(event)

        self.view.MustBeResized()

    def _initialize(self):

        wins = {
            "darwin": self._get_window_osx,
            "linux": self._get_window_linux,
            "win32": self._get_window_win,
        }

        self.view.SetWindow(wins.get(platform, self._get_window_linux)(self.winId()))

        self._initialized = True

    def _get_window_win(self, wid):

        from OCP.WNT import WNT_Window

        return WNT_Window(wid.ascapsule())

    def _get_window_linux(self, wid):

        from OCP.Xw import Xw_Window

        return Xw_Window(self.display_connection, int(wid))

    def _get_window_osx(self, wid):

        from OCP.Cocoa import Cocoa_Window

        return Cocoa_Window(wid.ascapsule())



from PyQt5.QtWidgets import QTreeWidget, QTreeWidgetItem, QAction
from PyQt5.QtCore import Qt, pyqtSlot, pyqtSignal

from OCP.AIS import AIS_ColoredShape
from OCP.gp import gp_Ax3

from cadquery import Vector

from ..mixins import ComponentMixin
from ..icons import icon


class CQChildItem(QTreeWidgetItem):

    def __init__(self, cq_item, **kwargs):

        super(CQChildItem, self).__init__(
            [type(cq_item).__name__, str(cq_item)], **kwargs
        )

        self.cq_item = cq_item


class CQStackItem(QTreeWidgetItem):

    def __init__(self, name, workplane=None, **kwargs):

        super(CQStackItem, self).__init__([name, ""], **kwargs)

        self.workplane = workplane


class CQObjectInspector(QTreeWidget, ComponentMixin):

    name = "CQ Object Inspector"

    sigRemoveObjects = pyqtSignal(list)
    sigDisplayObjects = pyqtSignal(list, bool)
    sigShowPlane = pyqtSignal([bool], [bool, float])
    sigChangePlane = pyqtSignal(gp_Ax3)

    def __init__(self, parent):

        super(CQObjectInspector, self).__init__(parent)
        self.setHeaderHidden(False)
        self.setRootIsDecorated(True)
        self.setContextMenuPolicy(Qt.ActionsContextMenu)
        self.setColumnCount(2)
        self.setHeaderLabels(["Type", "Value"])

        self.root = self.invisibleRootItem()
        self.inspected_items = []

        self._toolbar_actions = [
            QAction(
                icon("inspect"),
                "Inspect CQ object",
                self,
                toggled=self.inspect,
                checkable=True,
            )
        ]

        self.addActions(self._toolbar_actions)

    def menuActions(self):

        return {"Tools": self._toolbar_actions}

    def toolbarActions(self):

        return self._toolbar_actions

    @pyqtSlot(bool)
    def inspect(self, value):

        if value:
            self.itemSelectionChanged.connect(self.handleSelection)
            self.itemSelectionChanged.emit()
        else:
            self.itemSelectionChanged.disconnect(self.handleSelection)
            self.sigRemoveObjects.emit(self.inspected_items)
            self.sigShowPlane.emit(False)

    @pyqtSlot()
    def handleSelection(self):

        inspected_items = self.inspected_items
        self.sigRemoveObjects.emit(inspected_items)
        inspected_items.clear()

        items = self.selectedItems()
        if len(items) == 0:
            return

        item = items[-1]
        if type(item) is CQStackItem:
            cq_plane = item.workplane.plane
            dim = item.workplane.largestDimension()
            plane = gp_Ax3(
                cq_plane.origin.toPnt(), cq_plane.zDir.toDir(), cq_plane.xDir.toDir()
            )
            self.sigChangePlane.emit(plane)
            self.sigShowPlane[bool, float].emit(True, dim)

            for child in (item.child(i) for i in range(item.childCount())):
                obj = child.cq_item
                if hasattr(obj, "wrapped") and type(obj) != Vector:
                    ais = AIS_ColoredShape(obj.wrapped)
                    inspected_items.append(ais)

        else:
            self.sigShowPlane.emit(False)
            obj = item.cq_item
            if hasattr(obj, "wrapped") and type(obj) != Vector:
                ais = AIS_ColoredShape(obj.wrapped)
                inspected_items.append(ais)

        self.sigDisplayObjects.emit(inspected_items, False)

    @pyqtSlot(object)
    def setObject(self, cq_obj):

        self.root.takeChildren()

        # iterate through parent objects if they exist
        while getattr(cq_obj, "parent", None):
            current_frame = CQStackItem(str(cq_obj.plane.origin), workplane=cq_obj)
            self.root.addChild(current_frame)

            for obj in cq_obj.objects:
                current_frame.addChild(CQChildItem(obj))

            cq_obj = cq_obj.parent



import os

# import spyder.utils.encoding
from modulefinder import ModuleFinder

from .code_editor import CodeEditor

from PyQt5.QtCore import pyqtSignal, QFileSystemWatcher, QTimer, Qt, QEvent
from PyQt5.QtWidgets import (
    QAction,
    QFileDialog,
    QApplication,
    QListWidget,
    QListWidgetItem,
    QShortcut,
)
from PyQt5.QtGui import (
    QFontDatabase,
    QTextCursor,
    QKeyEvent,
)
from path import Path

import sys

import jedi

from pyqtgraph.parametertree import Parameter

from ..mixins import ComponentMixin
from ..utils import get_save_filename, get_open_filename, confirm
from .pyhighlight import PythonHighlighter

from ..icons import icon


class EditorDebugger:
    def __init__(self):
        self.breakpoints = []

    def get_breakpoints(self):
        return self.breakpoints

    def set_breakpoints(self, breakpoints):
        self.breakpoints = breakpoints
        return True


class Editor(CodeEditor, ComponentMixin):

    name = "Code Editor"

    # This signal is emitted whenever the currently-open file changes and
    # autoreload is enabled.
    triggerRerender = pyqtSignal(bool)
    sigFilenameChanged = pyqtSignal(str)
    statusChanged = pyqtSignal(str)

    preferences = Parameter.create(
        name="Preferences",
        children=[
            {"name": "Font size", "type": "int", "value": 12},
            {"name": "Autoreload", "type": "bool", "value": False},
            {"name": "Autoreload delay", "type": "int", "value": 50},
            {
                "name": "Autoreload: watch imported modules",
                "type": "bool",
                "value": False,
            },
            {"name": "Line wrap", "type": "bool", "value": False},
            {
                "name": "Color scheme",
                "type": "list",
                "values": ["Light", "Dark"],
                "value": "Light",
            },
            {"name": "Maximum line length", "type": "int", "value": 79},
        ],
    )

    EXTENSIONS = "py"

    # Tracks whether or not the document was saved from the internal editor vs an external editor
    was_modified_by_self = False

    # Helps display the completion list for the editor
    completion_list = None

    def __init__(self, parent=None):

        self._watched_file = None

        self.debugger = EditorDebugger()

        super(Editor, self).__init__(parent)
        ComponentMixin.__init__(self)

        self.setup_editor(
            line_numbers=True,
            markers=True,
            edge_line=self.preferences["Maximum line length"],
            tab_mode=False,
            show_blanks=True,
            font=QFontDatabase.systemFont(QFontDatabase.FixedFont),
            language="Python",
            filename="",
        )

        self._actions = {
            "File": [
                QAction(
                    icon("new"), "New", self, shortcut="ctrl+N", triggered=self.new
                ),
                QAction(
                    icon("open"), "Open", self, shortcut="ctrl+O", triggered=self.open
                ),
                QAction(
                    icon("save"), "Save", self, shortcut="ctrl+S", triggered=self.save
                ),
                QAction(
                    icon("save_as"),
                    "Save as",
                    self,
                    shortcut="ctrl+shift+S",
                    triggered=self.save_as,
                ),
                QAction(
                    icon("autoreload"),
                    "Automatic reload and preview",
                    self,
                    triggered=self.autoreload,
                    checkable=True,
                    checked=False,
                    objectName="autoreload",
                ),
            ]
        }

        for a in self._actions.values():
            self.addActions(a)

        self._fixContextMenu()

        # autoreload support
        self._file_watcher = QFileSystemWatcher(self)
        # we wait for 50ms after a file change for the file to be written completely
        self._file_watch_timer = QTimer(self)
        self._file_watch_timer.setInterval(self.preferences["Autoreload delay"])
        self._file_watch_timer.setSingleShot(True)
        self._file_watcher.fileChanged.connect(
            lambda val: self._file_watch_timer.start()
        )
        self._file_watch_timer.timeout.connect(self._file_changed)

        self.updatePreferences()

        # Create a floating list widget for completions
        self.completion_list = QListWidget(self)
        self.completion_list.setWindowFlags(Qt.Popup)
        self.completion_list.setFocusPolicy(Qt.NoFocus)
        self.completion_list.hide()

        # Connect the completion list to the editor
        self.completion_list.itemClicked.connect(self.insert_completion)

        # Ensure that when the escape key is pressed with the completion_list in focus, it will be hidden
        self.completion_list.installEventFilter(self)

        self.highlighter = PythonHighlighter(self.document())

    def eventFilter(self, watched, event):
        """
        Allows us to do things like escape and tab key press for the completion list.
        """

        if watched == self.completion_list and event.type() == QEvent.KeyPress:
            key_event = QKeyEvent(event)
            # Handle the escape key press
            if key_event.key() == Qt.Key_Escape:
                if self.completion_list and self.completion_list.isVisible():
                    self.completion_list.hide()
                    return True  # Event handled
            # Handle the tab key press
            elif key_event.key() == Qt.Key_Tab:
                if self.completion_list and self.completion_list.isVisible():
                    self.insert_completion(self.completion_list.currentItem())
                    return True  # Event handled
            elif key_event.key() == Qt.Key_Return:
                if self.completion_list and self.completion_list.isVisible():
                    self.insert_completion(self.completion_list.currentItem())
                    return True  # Event handled

            # Let the event propagate to the editor
            return False

        # Let the event propagate to the editor
        return False

    def _fixContextMenu(self):

        menu = self.menu

        # menu.removeAction(self.run_cell_action)
        # menu.removeAction(self.run_cell_and_advance_action)
        # menu.removeAction(self.run_selection_action)
        # menu.removeAction(self.re_run_last_cell_action)

    def updatePreferences(self, *args):

        self.set_color_scheme(self.preferences["Color scheme"])

        font = self.font
        font.setPointSize(self.preferences["Font size"])
        self.set_font(font)

        self.findChild(QAction, "autoreload").setChecked(self.preferences["Autoreload"])

        self._file_watch_timer.setInterval(self.preferences["Autoreload delay"])

        self.toggle_wrap_mode(self.preferences["Line wrap"])

        # Update the edge line (maximum line length)
        self.edge_line.set_enabled(True)
        self.edge_line.set_columns(self.preferences["Maximum line length"])

        self._clear_watched_paths()
        self._watch_paths()

    def confirm_discard(self):

        if self.modified:
            rv = confirm(
                self,
                "Please confirm",
                "Current document is not saved - do you want to continue?",
            )
        else:
            rv = True

        return rv

    def new(self):

        if not self.confirm_discard():
            return

        self.set_text("")
        self.filename = ""
        self.reset_modified()

    def open(self):

        if not self.confirm_discard():
            return

        curr_dir = Path(self.filename).absolute().dirname()
        fname = get_open_filename(self.EXTENSIONS, curr_dir)
        if fname != "":
            self.load_from_file(fname)

    def load_from_file(self, fname):

        self.set_text_from_file(fname)
        self.filename = fname
        self.reset_modified()

    def save(self):
        """
        Saves the current document to the current filename if it exists, otherwise it triggers a
        save-as dialog.
        """

        if self._filename != "":
            with open(self._filename, "w", encoding="utf-8") as f:
                f.write(self.toPlainText())

            # Let the editor and the rest of the app know that the file is no longer dirty
            self.reset_modified()

            self.was_modified_by_self = True

        else:
            self.save_as()

    def save_as(self):

        fname = get_save_filename(self.EXTENSIONS)
        if fname != "":
            with open(fname, "w", encoding="utf-8") as f:
                f.write(self.toPlainText())
                self.filename = fname

            self.reset_modified()

    def toggle_comment(self):
        """
        Allows us to mark the document as modified when the user toggles a comment.
        """
        super(Editor, self).toggle_comment()
        self.document().setModified(True)

    def _update_filewatcher(self):
        if self._watched_file and (
            self._watched_file != self.filename or not self.preferences["Autoreload"]
        ):
            self._clear_watched_paths()
            self._watched_file = None
        if (
            self.preferences["Autoreload"]
            and self.filename
            and self.filename != self._watched_file
        ):
            self._watched_file = self._filename
            self._watch_paths()

    @property
    def filename(self):
        return self._filename

    @filename.setter
    def filename(self, fname):
        self._filename = fname
        self._update_filewatcher()
        self.sigFilenameChanged.emit(fname)

    def _clear_watched_paths(self):
        paths = self._file_watcher.files()
        if paths:
            self._file_watcher.removePaths(paths)

    def _watch_paths(self):
        if Path(self._filename).exists():
            self._file_watcher.addPath(self._filename)
            if self.preferences["Autoreload: watch imported modules"]:
                module_paths = self.get_imported_module_paths(self._filename)
                if module_paths:
                    self._file_watcher.addPaths(module_paths)

    def _trigger_autocomplete(self):
        """
        Allows the user to ask for autocomplete suggestions.
        """

        # Clear the status bar
        self.statusChanged.emit("")

        # Track whether or not there are any completions to show
        completions_present = False

        script = jedi.Script(self.toPlainText(), path=self.filename)

        # Clear the completion list
        self.completion_list.clear()

        # Check to see if the character before the cursor is an open parenthesis
        cursor_pos = self.textCursor().position()
        text_before_cursor = self.toPlainText()[:cursor_pos]
        text_after_cursor = self.toPlainText()[cursor_pos:]
        if text_before_cursor.endswith("("):
            # If there is a trailing close parentheis after the cursor, remove it
            if text_after_cursor.startswith(")"):
                self.textCursor().deleteChar()

                # Update the script with the modified text
                script = jedi.Script(self.toPlainText(), path=self.filename)

            # Check if there are any function signatures
            signatures = script.get_signatures()
            if signatures:
                # Let the rest of the code know that there was a completion
                completions_present = True

                # Load the signatures into the completion list
                for signature in signatures:
                    # Build a human-readable signature
                    i = 0
                    cur_signature = f"{signature.name}("
                    for param in signature.params:
                        # Prevent trailing comma in parameter list
                        param_ending = ","
                        if i == len(signature.params) - 1:
                            param_ending = ""

                        # If the parameter is optional, do not overload the user with it
                        if "Optional" in param.description:
                            i += 1
                            continue

                        if "=" in param.description:
                            cur_signature += f"{param.name}={param.description.split('=')[1].strip()}{param_ending}"
                        else:
                            cur_signature += f"{param.name}{param_ending}"
                        i += 1
                    cur_signature += ")"

                    # Add the current signature to the list
                    item = QListWidgetItem(cur_signature)
                    self.completion_list.addItem(item)
        else:
            completions = script.complete()
            if completions:
                # Let the rest of the code know that there was a completion
                completions_present = True

                # Add completions to the list
                for completion in completions:
                    item = QListWidgetItem(completion.name)
                    self.completion_list.addItem(item)

        # Only show the completions list if there were any
        if completions_present:
            # Position the list near the cursor
            cursor_rect = self.cursorRect()
            global_pos = self.mapToGlobal(cursor_rect.bottomLeft())
            self.completion_list.move(global_pos)

            # Show the completion list
            self.completion_list.show()

            # Select the first item in the list
            self.completion_list.setCurrentRow(0)
        else:
            # Let the user know that no completions are available
            self.statusChanged.emit("No completions available")

    def insert_completion(self, item):
        """
        Inserts the selected completion into the editor.
        """

        # If there is an open parenthesis before the cursor, replace it with the completion
        if (
            self.textCursor().position() > 0
            and self.toPlainText()[self.textCursor().position() - 1] == "("
        ):
            cursor = self.textCursor()
            cursor.setPosition(cursor.position() - 1)
            cursor.movePosition(QTextCursor.EndOfWord, QTextCursor.KeepAnchor)
            cursor.removeSelectedText()

        # Find the last period in the text
        text_before_cursor = self.toPlainText()[: self.textCursor().position()]
        last_period_index = text_before_cursor.rfind(".")

        # Move the cursor to just after the last period position
        cursor = self.textCursor()
        cursor.setPosition(last_period_index + 1)

        # Remove text after last period
        cursor.movePosition(QTextCursor.EndOfWord, QTextCursor.KeepAnchor)
        cursor.removeSelectedText()

        # Insert the completion text
        cursor.insertText(item.text())
        self.setTextCursor(cursor)

        # Hide the completion list
        self.completion_list.hide()

    # callback triggered by QFileSystemWatcher
    def _file_changed(self):
        # neovim writes a file by removing it first so must re-add each time
        self._watch_paths()

        # Save the current cursor position and selection
        cursor = self.textCursor()
        cursor_position = cursor.position()
        anchor_position = cursor.anchor()

        # Save the current scroll position
        vertical_scroll_pos = self.verticalScrollBar().value()
        horizontal_scroll_pos = self.horizontalScrollBar().value()

        # Block signals to avoid reset issues
        self.blockSignals(True)

        # Read the contents of the file into a string
        with open(self._filename, "r", encoding="utf-8") as f:
            file_contents = f.read()

            # Insert new text while preserving history
            cursor = self.textCursor()
            cursor.select(QTextCursor.Document)
            cursor.insertText(file_contents)

            # The editor will not always update after a text insertion, so we force it
            QApplication.processEvents()

        # Stop blocking signals
        self.blockSignals(False)

        self.document().setModified(True)

        # Undo has to be backed up one step to compensate for the text insertion
        if self.was_modified_by_self:
            self.document().undo()
            self.was_modified_by_self = False

        # Restore the cursor position and selection
        cursor.setPosition(anchor_position)
        cursor.setPosition(cursor_position, QTextCursor.KeepAnchor)
        self.setTextCursor(cursor)

        # Restore the scroll position
        self.verticalScrollBar().setValue(vertical_scroll_pos)
        self.horizontalScrollBar().setValue(horizontal_scroll_pos)

        # Reset the dirty state and trigger a 3D render
        self.reset_modified()
        self.triggerRerender.emit(True)

    # Turn autoreload on/off.
    def autoreload(self, enabled):
        self.preferences["Autoreload"] = enabled
        self._update_filewatcher()

    def reset_modified(self):

        self.document().setModified(False)

    @property
    def modified(self):

        return self.document().isModified()

    def saveComponentState(self, store):

        if self.filename != "":
            store.setValue(self.name + "/state", self.filename)

    def restoreComponentState(self, store):

        filename = store.value(self.name + "/state")

        if filename and self.filename == "":
            try:
                self.load_from_file(filename)
            except IOError:
                self._logger.warning(f"could not open {filename}")

    def get_imported_module_paths(self, module_path):

        finder = ModuleFinder([os.path.dirname(module_path)])
        imported_modules = []

        try:
            finder.run_script(module_path)
        except SyntaxError as err:
            self._logger.warning(f"Syntax error in {module_path}: {err}")
        except Exception as err:
            # The module finder has trouble when CadQuery is imported in the top level script and in
            # imported modules. The warning about it can be ignored.
            if "cadquery" not in finder.badmodules or (
                "cadquery" in finder.badmodules and len(finder.badmodules) > 1
            ):
                self._logger.warning(
                    f"Cannot determine imported modules in {module_path}: {type(err).__name__} {err}"
                )
        else:
            for module_name, module in finder.modules.items():
                if module_name != "__main__":
                    path = getattr(module, "__file__", None)
                    if path is not None and os.path.isfile(path):
                        imported_modules.append(path)

        return imported_modules


if __name__ == "__main__":

    from PyQt5.QtWidgets import QApplication

    app = QApplication(sys.argv)
    editor = Editor()
    editor.show()

    sys.exit(app.exec_())



import os, sys, asyncio

if "CASROOT" in os.environ:
    del os.environ["CASROOT"]

if sys.platform == "win32":
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

from cq_editor.__main__ import main


if __name__ == "__main__":
    main()



from PyQt5.QtWidgets import QTreeWidget, QTreeWidgetItem, QStackedWidget, QDialog
from PyQt5.QtCore import pyqtSlot, Qt

from pyqtgraph.parametertree import ParameterTree

from .utils import splitter, layout


class PreferencesTreeItem(QTreeWidgetItem):

    def __init__(
        self,
        name,
        widget,
    ):

        super(PreferencesTreeItem, self).__init__(name)
        self.widget = widget


class PreferencesWidget(QDialog):

    def __init__(self, parent, components):

        super(PreferencesWidget, self).__init__(
            parent,
            Qt.Window | Qt.CustomizeWindowHint | Qt.WindowCloseButtonHint,
            windowTitle="Preferences",
        )

        self.stacked = QStackedWidget(self)
        self.preferences_tree = QTreeWidget(
            self,
            headerHidden=True,
            itemsExpandable=False,
            rootIsDecorated=False,
            columnCount=1,
        )

        self.root = self.preferences_tree.invisibleRootItem()

        self.add("General", parent)

        for v in parent.components.values():
            self.add(v.name, v)

        self.splitter = splitter((self.preferences_tree, self.stacked), (2, 5))
        layout(self, (self.splitter,), self)

        self.preferences_tree.currentItemChanged.connect(self.handleSelection)

    def add(self, name, component):

        if component.preferences:
            widget = ParameterTree()
            widget.setHeaderHidden(True)
            widget.setParameters(component.preferences, showTop=False)
            self.root.addChild(PreferencesTreeItem((name,), widget))

            self.stacked.addWidget(widget)

            # PyQtGraph is not setting items in drop down lists properly, so we do it manually
            for child in component.preferences.children():
                # Fill the editor color scheme drop down list
                if child.name() == "Color scheme":
                    child.setLimits(["Light", "Dark"])
                # Fill the camera projection type
                elif child.name() == "Projection Type":
                    child.setLimits(
                        [
                            "Orthographic",
                            "Perspective",
                            "Stereo",
                            "MonoLeftEye",
                            "MonoRightEye",
                        ]
                    )
                # Fill the stereo mode, or lack thereof
                elif child.name() == "Stereo Mode":
                    child.setLimits(
                        [
                            "QuadBuffer",
                            "Anaglyph",
                            "RowInterlaced",
                            "ColumnInterlaced",
                            "ChessBoard",
                            "SideBySide",
                            "OverUnder",
                        ]
                    )
                # Fill the light/dark theme in the general settings
                elif child.name() == "Light/Dark Theme":
                    child.setLimits(["Light", "Dark"])
                # Fill the orbit method
                elif child.name() == "Orbit Method":
                    child.setLimits(
                        [
                            "Turntable",
                            "Trackball",
                        ]
                    )

    @pyqtSlot(QTreeWidgetItem, QTreeWidgetItem)
    def handleSelection(self, item, *args):

        if item:
            self.stacked.setCurrentWidget(item.widget)



import sys

from PyQt5.QtCore import QObject, pyqtSignal
from PyQt5.QtGui import QPalette, QColor
from PyQt5.QtWidgets import (
    QLabel,
    QMainWindow,
    QToolBar,
    QDockWidget,
    QAction,
    QApplication,
    QMenu,
)
from logbook import Logger
import cadquery as cq

from .widgets.editor import Editor
from .widgets.viewer import OCCViewer
from .widgets.console import ConsoleWidget
from .widgets.object_tree import ObjectTree
from .widgets.traceback_viewer import TracebackPane
from .widgets.debugger import Debugger, LocalsView
from .widgets.cq_object_inspector import CQObjectInspector
from .widgets.log import LogViewer

from . import __version__
from .utils import (
    dock,
    add_actions,
    open_url,
    about_dialog,
    check_gtihub_for_updates,
    confirm,
)
from .mixins import MainMixin
from .icons import icon
from pyqtgraph.parametertree import Parameter
from .preferences import PreferencesWidget


class _PrintRedirectorSingleton(QObject):
    """This class monkey-patches `sys.stdout.write` to emit a signal.
    It is instanciated as `.main_window.PRINT_REDIRECTOR` and should not be instanciated again.
    """

    sigStdoutWrite = pyqtSignal(str)

    def __init__(self):
        super().__init__()

        original_stdout_write = sys.stdout.write

        def new_stdout_write(text: str):
            self.sigStdoutWrite.emit(text)
            return original_stdout_write(text)

        sys.stdout.write = new_stdout_write


PRINT_REDIRECTOR = _PrintRedirectorSingleton()


class MainWindow(QMainWindow, MainMixin):

    name = "CQ-Editor"
    org = "CadQuery"

    preferences = Parameter.create(
        name="Preferences",
        children=[
            {
                "name": "Light/Dark Theme",
                "type": "list",
                "value": "Light",
                "values": [
                    "Light",
                    "Dark",
                ],
            },
        ],
    )

    def __init__(self, parent=None, filename=None):

        super(MainWindow, self).__init__(parent)
        MainMixin.__init__(self)

        self.setWindowIcon(icon("app"))

        # Windows workaround - makes the correct task bar icon show up.
        if sys.platform == "win32":
            import ctypes

            myappid = "cq-editor"  # arbitrary string
            ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)

        self.viewer = OCCViewer(self)
        self.setCentralWidget(self.viewer.canvas)

        self.prepare_panes()
        self.registerComponent("viewer", self.viewer)
        self.prepare_toolbar()
        self.prepare_menubar()

        self.prepare_statusbar()
        self.prepare_actions()

        self.components["object_tree"].addLines()

        self.prepare_console()

        self.fill_dummy()

        self.setup_logging()

        # Allows us to react to the top-level settings for this window being changed
        self.preferences.sigTreeStateChanged.connect(self.preferencesChanged)

        self.restorePreferences()
        self.restoreWindow()

        # Handle the event of the editor being hidden or shown
        self.editor_dock = self.docks["editor"]
        self.editor_dock.visibilityChanged.connect(self.handleEditorVisiblityChange)
        self.handleEditorVisiblityChange(not self.editor_dock.isHidden())

        # Let the user know when the file has been modified
        self.components["editor"].document().modificationChanged.connect(
            self.update_window_title
        )

        if filename:
            self.components["editor"].load_from_file(filename)

        self.restoreComponentState()

    def handleEditorVisiblityChange(self, visible):
        """
        Does the work required to enable/disable menu items when the Editor visibility is changed.
        """
        self.toggle_comment_action.setEnabled(visible)
        self.autocomplete_action.setEnabled(visible)

    def preferencesChanged(self, param, changes):
        """
        Triggered when the preferences for this window are changed.
        """

        # Use the default light theme/palette
        if self.preferences["Light/Dark Theme"] == "Light":
            QApplication.instance().setStyleSheet("")
            QApplication.instance().setPalette(QApplication.style().standardPalette())

            # The console theme needs to be changed separately
            self.components["console"].app_theme_changed("Light")
        # Use the dark theme/palette
        elif self.preferences["Light/Dark Theme"] == "Dark":
            QApplication.instance().setStyle("Fusion")

            # Now use a palette to switch to dark colors:
            white_color = QColor(255, 255, 255)
            black_color = QColor(0, 0, 0)
            red_color = QColor(255, 0, 0)
            palette = QPalette()
            palette.setColor(QPalette.Window, QColor(53, 53, 53))
            palette.setColor(QPalette.WindowText, white_color)
            palette.setColor(QPalette.Base, QColor(25, 25, 25))
            palette.setColor(QPalette.AlternateBase, QColor(53, 53, 53))
            palette.setColor(QPalette.ToolTipBase, black_color)
            palette.setColor(QPalette.ToolTipText, white_color)
            palette.setColor(QPalette.Text, white_color)
            palette.setColor(QPalette.Button, QColor(53, 53, 53))
            palette.setColor(QPalette.ButtonText, white_color)
            palette.setColor(QPalette.BrightText, red_color)
            palette.setColor(QPalette.Link, QColor(42, 130, 218))
            palette.setColor(QPalette.Highlight, QColor(42, 130, 218))
            palette.setColor(QPalette.HighlightedText, black_color)
            QApplication.instance().setPalette(palette)

            # The console theme needs to be changed separately
            self.components["console"].app_theme_changed("Dark")

        # We alter the color of the toolbar separately to avoid having separate dark theme icons
        p = self.toolbar.palette()
        if self.preferences["Light/Dark Theme"] == "Dark":
            p.setColor(QPalette.Background, QColor(120, 120, 120))

            # TWeak the QMenu items palette for dark theme
            menu_palette = self.menuBar().palette()
            menu_palette.setColor(QPalette.Base, QColor(80, 80, 80))
            for menu in self.menuBar().findChildren(QMenu):
                menu.setPalette(menu_palette)
        else:
            p.setColor(QPalette.Background, QColor(240, 240, 240))

            # Revert the QMenu items palette for dark theme
            menu_palette = self.menuBar().palette()
            menu_palette.setColor(QPalette.Base, QColor(240, 240, 240))
            for menu in self.menuBar().findChildren(QMenu):
                menu.setPalette(menu_palette)

        self.toolbar.setPalette(p)

    def closeEvent(self, event):

        self.saveWindow()
        self.savePreferences()
        self.saveComponentState()

        if self.components["editor"].document().isModified():

            rv = confirm(self, "Confirm close", "Close without saving?")

            if rv:
                event.accept()
                super(MainWindow, self).closeEvent(event)
            else:
                event.ignore()
        else:
            super(MainWindow, self).closeEvent(event)

    def prepare_panes(self):

        self.registerComponent(
            "editor",
            Editor(self),
            lambda c: dock(c, "Editor", self, defaultArea="left"),
        )

        self.registerComponent(
            "object_tree",
            ObjectTree(self),
            lambda c: dock(c, "Objects", self, defaultArea="right"),
        )

        self.registerComponent(
            "traceback_viewer",
            TracebackPane(self),
            lambda c: dock(c, "Current traceback", self, defaultArea="bottom"),
        )

        self.registerComponent("debugger", Debugger(self))

        self.registerComponent(
            "console",
            ConsoleWidget(self),
            lambda c: dock(c, "Console", self, defaultArea="bottom"),
        )

        self.registerComponent(
            "variables_viewer",
            LocalsView(self),
            lambda c: dock(c, "Variables", self, defaultArea="right"),
        )

        self.registerComponent(
            "cq_object_inspector",
            CQObjectInspector(self),
            lambda c: dock(c, "CQ object inspector", self, defaultArea="right"),
        )
        self.registerComponent(
            "log",
            LogViewer(self),
            lambda c: dock(c, "Log viewer", self, defaultArea="bottom"),
        )

        for d in self.docks.values():
            d.show()

        PRINT_REDIRECTOR.sigStdoutWrite.connect(
            lambda text: self.components["log"].append(text)
        )

    def prepare_menubar(self):

        menu = self.menuBar()

        menu_file = menu.addMenu("&File")
        menu_edit = menu.addMenu("&Edit")
        menu_tools = menu.addMenu("&Tools")
        menu_run = menu.addMenu("&Run")
        menu_view = menu.addMenu("&View")
        menu_help = menu.addMenu("&Help")

        # per component menu elements
        menus = {
            "File": menu_file,
            "Edit": menu_edit,
            "Run": menu_run,
            "Tools": menu_tools,
            "View": menu_view,
            "Help": menu_help,
        }

        for comp in self.components.values():
            self.prepare_menubar_component(menus, comp.menuActions())

        # global menu elements
        menu_view.addSeparator()
        for d in self.findChildren(QDockWidget):
            menu_view.addAction(d.toggleViewAction())

        menu_view.addSeparator()
        for t in self.findChildren(QToolBar):
            menu_view.addAction(t.toggleViewAction())

        self.toggle_comment_action = QAction(
            icon("toggle-comment"),
            "Toggle Comment",
            self,
            shortcut="ctrl+/",
            triggered=self.components["editor"].toggle_comment,
        )
        menu_edit.addAction(self.toggle_comment_action)

        # Add the menu action to toggle auto-completion
        self.autocomplete_action = QAction(
            icon("search"),
            "Auto-Complete",
            self,
            shortcut="alt+/",
            triggered=self.components["editor"]._trigger_autocomplete,
        )
        menu_edit.addAction(self.autocomplete_action)

        # Add the menu action to open the code search controls
        self.search_action = QAction(
            icon("search"),
            "Search",
            self,
            shortcut="ctrl+F",
            triggered=self.components["editor"].search_widget.show_search,
        )
        menu_edit.addAction(self.search_action)

        menu_edit.addAction(
            QAction(
                icon("preferences"),
                "Preferences",
                self,
                triggered=self.edit_preferences,
            )
        )

        menu_help.addAction(
            QAction(icon("help"), "Documentation", self, triggered=self.documentation)
        )

        menu_help.addAction(
            QAction("CQ documentation", self, triggered=self.cq_documentation)
        )

        menu_help.addAction(QAction(icon("about"), "About", self, triggered=self.about))

        menu_help.addAction(
            QAction(
                "Check for CadQuery updates", self, triggered=self.check_for_cq_updates
            )
        )

    def prepare_menubar_component(self, menus, comp_menu_dict):

        for name, action in comp_menu_dict.items():
            menus[name].addActions(action)

    def prepare_toolbar(self):

        self.toolbar = QToolBar("Main toolbar", self, objectName="Main toolbar")

        for c in self.components.values():
            add_actions(self.toolbar, c.toolbarActions())

        self.addToolBar(self.toolbar)

    def prepare_statusbar(self):

        self.status_label = QLabel("", parent=self)
        self.statusBar().insertPermanentWidget(0, self.status_label)

    def prepare_actions(self):

        self.components["debugger"].sigRendered.connect(
            self.components["object_tree"].addObjects
        )
        self.components["debugger"].sigTraceback.connect(
            self.components["traceback_viewer"].addTraceback
        )
        self.components["debugger"].sigLocals.connect(
            self.components["variables_viewer"].update_frame
        )
        self.components["debugger"].sigLocals.connect(
            self.components["console"].push_vars
        )

        self.components["object_tree"].sigObjectsAdded[list].connect(
            self.components["viewer"].display_many
        )
        self.components["object_tree"].sigObjectsAdded[list, bool].connect(
            self.components["viewer"].display_many
        )
        self.components["object_tree"].sigItemChanged.connect(
            self.components["viewer"].update_item
        )
        self.components["object_tree"].sigObjectsRemoved.connect(
            self.components["viewer"].remove_items
        )
        self.components["object_tree"].sigCQObjectSelected.connect(
            self.components["cq_object_inspector"].setObject
        )
        self.components["object_tree"].sigObjectPropertiesChanged.connect(
            self.components["viewer"].redraw
        )
        self.components["object_tree"].sigAISObjectsSelected.connect(
            self.components["viewer"].set_selected
        )

        self.components["viewer"].sigObjectSelected.connect(
            self.components["object_tree"].handleGraphicalSelection
        )

        self.components["traceback_viewer"].sigHighlightLine.connect(
            self.components["editor"].go_to_line
        )

        self.components["cq_object_inspector"].sigDisplayObjects.connect(
            self.components["viewer"].display_many
        )
        self.components["cq_object_inspector"].sigRemoveObjects.connect(
            self.components["viewer"].remove_items
        )
        self.components["cq_object_inspector"].sigShowPlane.connect(
            self.components["viewer"].toggle_grid
        )
        self.components["cq_object_inspector"].sigShowPlane[bool, float].connect(
            self.components["viewer"].toggle_grid
        )
        self.components["cq_object_inspector"].sigChangePlane.connect(
            self.components["viewer"].set_grid_orientation
        )

        self.components["debugger"].sigLocalsChanged.connect(
            self.components["variables_viewer"].update_frame
        )
        self.components["debugger"].sigLineChanged.connect(
            self.components["editor"].go_to_line
        )
        self.components["debugger"].sigDebugging.connect(
            self.components["object_tree"].stashObjects
        )
        self.components["debugger"].sigCQChanged.connect(
            self.components["object_tree"].addObjects
        )
        self.components["debugger"].sigTraceback.connect(
            self.components["traceback_viewer"].addTraceback
        )

        # trigger re-render when file is modified externally or saved
        self.components["editor"].triggerRerender.connect(
            self.components["debugger"].render
        )
        self.components["editor"].sigFilenameChanged.connect(
            self.handle_filename_change
        )
        # Allows updating of the status bar from the Editor
        self.components["editor"].statusChanged.connect(self.update_statusbar)

    def prepare_console(self):

        console = self.components["console"]
        obj_tree = self.components["object_tree"]

        # application related items
        console.push_vars({"self": self})

        # CQ related items
        console.push_vars(
            {
                "show": obj_tree.addObject,
                "show_object": obj_tree.addObject,
                "rand_color": self.components["debugger"]._rand_color,
                "cq": cq,
                "log": Logger(self.name).info,
            }
        )

    def fill_dummy(self):

        self.components["editor"].set_text(
            'import cadquery as cq\nresult = cq.Workplane("XY" ).box(3, 3, 0.5).edges("|Z").fillet(0.125)\nshow_object(result)'
        )

    def setup_logging(self):

        from logbook.compat import redirect_logging
        from logbook import INFO, Logger

        redirect_logging()
        self.components["log"].handler.level = INFO
        self.components["log"].handler.push_application()

        self._logger = Logger(self.name)

        def handle_exception(exc_type, exc_value, exc_traceback):

            if issubclass(exc_type, KeyboardInterrupt):
                sys.__excepthook__(exc_type, exc_value, exc_traceback)
                return

            self._logger.error(
                "Uncaught exception occurred",
                exc_info=(exc_type, exc_value, exc_traceback),
            )

        sys.excepthook = handle_exception

    def edit_preferences(self):

        prefs = PreferencesWidget(self, self.components)
        prefs.exec_()

    def about(self):

        about_dialog(
            self,
            f"About CQ-editor",
            f"PyQt GUI for CadQuery.\nVersion: {__version__}.\nSource Code: https://github.com/CadQuery/CQ-editor",
        )

    def check_for_cq_updates(self):

        check_gtihub_for_updates(self, cq)

    def documentation(self):

        open_url("https://github.com/CadQuery")

    def cq_documentation(self):

        open_url("https://cadquery.readthedocs.io/en/latest/")

    def handle_filename_change(self, fname):

        new_title = fname if fname else "*"
        self.setWindowTitle(f"{self.name}: {new_title}")

    def update_window_title(self, modified):
        """
        Allows updating the window title to show that the document has been modified.
        """
        title = self.windowTitle().rstrip("*")
        if modified:
            title += "*"
        self.setWindowTitle(title)

    def update_statusbar(self, status_text):
        """
        Allow updating the status bar with information.
        """

        # Update the statusbar text
        self.status_label.setText(status_text)


if __name__ == "__main__":

    pass



import sys
import argparse

from PyQt5.QtWidgets import QApplication

NAME = "CQ-editor"

# need to initialize QApp here, otherewise svg icons do not work on windows
app = QApplication(sys.argv, applicationName=NAME)

from .main_window import MainWindow


def main():

    parser = argparse.ArgumentParser(description=NAME)
    parser.add_argument("filename", nargs="?", default=None)

    args = parser.parse_args(app.arguments()[1:])

    # sys.exit(app.exec_())

    try:
        win = MainWindow(filename=args.filename if args.filename else None)
        win.show()
        app.exec_()
    except Exception as e:
        import traceback

        traceback.print_exc()


if __name__ == "__main__":

    main()



# Much of this code was adapted from https://github.com/leixingyu/codeEditor which is under
# an MIT license
import os
from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtGui import QPalette, QColor


DARK_BLUE = QtGui.QColor(118, 150, 185)


class SearchWidget(QtWidgets.QWidget):
    def __init__(self, editor):
        super(SearchWidget, self).__init__(editor)
        self.editor = editor
        self.current_match = 0
        self.total_matches = 0

        self.setup_ui()

        # This widget should initially be hidden
        self.hide()

    def setup_ui(self):
        # Horizontal layout
        layout = QtWidgets.QHBoxLayout()
        layout.setContentsMargins(5, 5, 5, 5)

        # Search input box
        self.search_input = QtWidgets.QLineEdit()
        self.search_input.setPlaceholderText("Search...")
        self.search_input.setMinimumWidth(100)
        self.search_input.textChanged.connect(self.on_search_text_changed)
        self.search_input.returnPressed.connect(self.find_next)

        # Previous button
        self.prev_button = QtWidgets.QPushButton("Prev")
        self.prev_button.clicked.connect(self.find_previous)
        self.prev_button.setEnabled(False)

        # Next button
        self.next_button = QtWidgets.QPushButton("Next")
        self.next_button.clicked.connect(self.find_next)
        self.next_button.setEnabled(False)

        # Match count label
        self.match_label = QtWidgets.QLabel("0 matches")

        # Close button
        self.close_button = QtWidgets.QPushButton("×")
        self.close_button.setMaximumSize(20, 20)
        self.close_button.clicked.connect(self.hide_search)

        # Add widgets to layout
        layout.addWidget(self.search_input)
        layout.addWidget(self.prev_button)
        layout.addWidget(self.next_button)
        layout.addWidget(self.match_label)
        layout.addWidget(self.close_button)
        self.setLayout(layout)

    def on_search_text_changed(self, text):
        """
        Called as the user types text into the search field.
        """
        if not text:
            self.clear_highlights()
            self.update_match_count(0, 0)
            self.prev_button.setEnabled(False)
            self.next_button.setEnabled(False)
            return

        self.find_all_matches(text)

    def find_all_matches(self, search_text):
        """
        Finds all the matches within the search text.
        """
        if not search_text:
            return

        # Clear any previous highlights
        self.clear_highlights()

        # Find all matches
        document = self.editor.document()
        cursor = QtGui.QTextCursor(document)
        self.matches = []

        # Find all occurrences
        while True:
            # Look for a match
            cursor = document.find(search_text, cursor)
            if cursor.isNull():
                break
            self.matches.append(cursor)

        self.total_matches = len(self.matches)

        # If there are matches make them visible to the user
        if self.total_matches > 0:
            self.current_match = 0
            self.highlight_matches()
            self.highlight_current_match()
            self.prev_button.setEnabled(True)
            self.next_button.setEnabled(True)
        else:
            self.prev_button.setEnabled(False)
            self.next_button.setEnabled(False)

        self.update_match_count(
            self.current_match + 1 if self.total_matches > 0 else 0, self.total_matches
        )

    def highlight_matches(self):
        """
        Highlights all matches to make them visible.
        """
        extra_selections = []

        for cursor in self.matches:
            selection = QtWidgets.QTextEdit.ExtraSelection()
            selection.format.setBackground(QtGui.QColor(255, 255, 0, 100))
            selection.cursor = cursor
            extra_selections.append(selection)

        self.editor.setExtraSelections(extra_selections)

    def highlight_current_match(self):
        """
        Makes the current match stand out from the others.
        """

        # If there are no matches to highlight, then skip this step
        if not self.matches or self.current_match >= len(self.matches):
            return

        # Highlight current match more than others and scroll to it
        extra_selections = []

        for i, cursor in enumerate(self.matches):
            selection = QtWidgets.QTextEdit.ExtraSelection()
            # The current match should stand out
            if i == self.current_match:
                selection.format.setBackground(QtGui.QColor(255, 165, 0))
            else:
                selection.format.setBackground(QtGui.QColor(255, 255, 0, 100))
            selection.cursor = cursor
            extra_selections.append(selection)

        self.editor.setExtraSelections(extra_selections)

        # Scroll to the current match
        self.editor.setTextCursor(self.matches[self.current_match])
        self.editor.ensureCursorVisible()

    def find_next(self):
        """
        Finds the next match.
        """

        # If there are no matches, skip this step
        if not self.matches:
            return

        self.current_match = (self.current_match + 1) % len(self.matches)
        self.highlight_current_match()
        self.update_match_count(self.current_match + 1, self.total_matches)

    def find_previous(self):
        """
        Finds the previous match.
        """

        # If there are no matches, skip this step
        if not self.matches:
            return

        self.current_match = (self.current_match - 1) % len(self.matches)
        self.highlight_current_match()
        self.update_match_count(self.current_match + 1, self.total_matches)

    def update_match_count(self, current, total):
        """
        Updates the match count for the user.
        """
        if total == 0:
            self.match_label.setText("0 matches")
        else:
            self.match_label.setText(f"{current} of {total}")

    def clear_highlights(self):
        """
        Clears all of the find highlights.
        """
        self.editor.setExtraSelections([])
        self.matches = []

    def show_search(self):
        """
        Makes the search dialog visible.
        """
        self.show()

        # Make sure the user can start typing search text right away
        self.search_input.setFocus()
        self.search_input.selectAll()

        self.position_widget()

        # If there is already text in the search box, trigger the search
        if self.search_input.text():
            self.find_all_matches(self.search_input.text())

    def hide_search(self):
        """
        Hides the search dialog again.
        """
        self.hide()
        self.clear_highlights()
        self.editor.setFocus()

    def position_widget(self):
        """
        Makes sure that the search widget gets placed in the right location
        in the window.
        """

        # Top-right corner of the editor
        editor_rect = self.editor.geometry()
        widget_width = 400
        widget_height = 40
        x = editor_rect.width() - widget_width - 20
        y = 10

        # Set the size of the widget and bring it to the front
        self.setGeometry(x, y, widget_width, widget_height)
        self.raise_()


class LineNumberArea(QtWidgets.QWidget):
    def __init__(self, editor):
        super(LineNumberArea, self).__init__(editor)
        self._code_editor = editor

    def sizeHint(self):
        return QtCore.QSize(self._code_editor.line_number_area_width(), 0)

    def mousePressEvent(self, event):
        """
        Handles mouse clicks to add/remove breakpoints.
        """
        if event.button() == QtCore.Qt.LeftButton:
            # Calculate which line was clicked
            line_number = self.get_line_number_from_position(event.pos())
            if line_number is not None:
                self._code_editor.toggle_breakpoint(line_number)

    def get_line_number_from_position(self, pos):
        """
        Convert mouse position to line number.
        """

        # Get the first visible block
        block = self._code_editor.firstVisibleBlock()
        block_number = block.blockNumber()
        offset = self._code_editor.contentOffset()
        top = self._code_editor.blockBoundingGeometry(block).translated(offset).top()

        # Find which block the click position corresponds to
        while block.isValid():
            bottom = top + self._code_editor.blockBoundingRect(block).height()

            if top <= pos.y() <= bottom:
                return (
                    block_number + 1
                )  # Line numbers start at 0 internally and 1 for the user

            block = block.next()
            top = bottom
            block_number += 1

        return None

    def paintEvent(self, event):
        self._code_editor.lineNumberAreaPaintEvent(event)


class EdgeLine(QtWidgets.QWidget):
    edge_line = None
    columns = 80

    def __init__(self):
        super(QtWidgets.QWidget, self).__init__()

    def set_enabled(self, enabled_state):
        self.setEnabled = enabled_state

    def set_columns(self, number_of_columns):
        self.columns = number_of_columns


class CodeTextEdit(QtWidgets.QPlainTextEdit):
    is_first = False
    pressed_keys = list()

    indented = QtCore.pyqtSignal(object)
    unindented = QtCore.pyqtSignal(object)

    def __init__(self):
        super(CodeTextEdit, self).__init__()

        self.indented.connect(self.do_indent)
        self.unindented.connect(self.undo_indent)

    def clear_selection(self):
        """
        Clear text selection on cursor
        """
        cursor = self.textCursor()
        pos = cursor.selectionEnd()
        cursor.movePosition(pos)
        self.setTextCursor(cursor)

    def get_selection_range(self):
        """
        Get text selection line range from cursor
        Note: currently only support continuous selection

        :return: (int, int). start line number and end line number
        """
        cursor = self.textCursor()
        if not cursor.hasSelection():
            return 0, 0

        start_pos = cursor.selectionStart()
        end_pos = cursor.selectionEnd()

        cursor.setPosition(start_pos)
        start_line = cursor.blockNumber()
        cursor.setPosition(end_pos)
        end_line = cursor.blockNumber()

        return start_line, end_line

    def remove_line_start(self, string, line_number):
        """
        Remove certain string occurrence on line start

        :param string: str. string pattern to remove
        :param line_number: int. line number
        """
        cursor = QtGui.QTextCursor(self.document().findBlockByLineNumber(line_number))
        cursor.select(QtGui.QTextCursor.LineUnderCursor)
        text = cursor.selectedText()
        if text.startswith(string):
            cursor.removeSelectedText()
            cursor.insertText(text.split(string, 1)[-1])

    def insert_line_start(self, string, line_number):
        """
        Insert certain string pattern on line start

        :param string: str. string pattern to insert
        :param line_number: int. line number
        """
        cursor = QtGui.QTextCursor(self.document().findBlockByLineNumber(line_number))
        self.setTextCursor(cursor)
        self.textCursor().insertText(string)

    def keyPressEvent(self, event):
        """
        Extend the key press event to create key shortcuts
        """
        self.is_first = True
        self.pressed_keys.append(event.key())
        start_line, end_line = self.get_selection_range()

        # indent event
        if event.key() == QtCore.Qt.Key_Tab:
            lines = range(start_line, end_line + 1)
            self.indented.emit(lines)
            return
        elif event.key() == QtCore.Qt.Key_Tab and (end_line - start_line):
            lines = range(start_line, end_line + 1)
            self.indented.emit(lines)
            return
        # un-indent event
        elif event.key() == QtCore.Qt.Key_Backtab:
            lines = range(start_line, end_line + 1)
            self.unindented.emit(lines)
            return

        super(CodeTextEdit, self).keyPressEvent(event)

    def do_indent(self, lines):
        """
        Indent lines

        :param lines: [int]. line numbers
        """

        # Get the selection range of lines
        start_line, end_line = self.get_selection_range()

        # If a single line is selected, make sure it is the only line in the range
        if start_line == 0 and end_line == 0:
            # Insert a tab at the current cursor location
            cursor = self.textCursor()
            cursor.insertText("    ")

            # Make sure that no lines are changed
            lines = []
        # Multiple lines have been selected
        else:
            lines = range(start_line, end_line + 1)

        # Walk through the selected lines and tab them (with 4 spaces)
        for line in lines:
            self.insert_line_start("    ", line)

    def undo_indent(self, lines):
        """
        Un-indent lines

        :param lines: [int]. line numbers
        """
        for line in lines:
            self.remove_line_start("    ", line)

        # Set the cursor to the beginning of the last line
        cursor = self.textCursor()
        cursor.setPosition(cursor.selectionEnd())  # Move to end of selection
        cursor.movePosition(QtGui.QTextCursor.StartOfLine)  # Jump to start of line
        self.setTextCursor(cursor)


class CodeEditor(CodeTextEdit):
    def __init__(self, parent=None):
        super(CodeEditor, self).__init__()
        self.line_number_area = LineNumberArea(self)

        self.font = QtGui.QFont()
        self.font.setFamily("Courier New")
        self.font.setStyleHint(QtGui.QFont.Monospace)
        self.font.setPointSize(10)
        self.setFont(self.font)

        self.tab_size = 4
        self.setTabStopWidth(self.tab_size * self.fontMetrics().width(" "))

        self.blockCountChanged.connect(self.update_line_number_area_width)
        self.updateRequest.connect(self.update_line_number_area)
        # self.cursorPositionChanged.connect(self.highlight_current_line)

        self.update_line_number_area_width(0)
        # self.highlight_current_line()

        self.menu = QtWidgets.QMenu()

        self.edge_line = EdgeLine()

        self.search_widget = SearchWidget(self)

        self._filename = ""

    def keyPressEvent(self, event):
        # Handle Ctrl+F for search
        if (
            event.modifiers() == QtCore.Qt.ControlModifier
            and event.key() == QtCore.Qt.Key_F
        ):
            self.search_widget.show_search()
            return

        # Handle F3 for find next (when search widget is visible)
        if event.key() == QtCore.Qt.Key_F3 and self.search_widget.isVisible():
            if event.modifiers() == QtCore.Qt.AltModifier:
                self.search_widget.find_previous()  # Alt+F3 for previous
            else:
                self.search_widget.find_next()  # F3 for next
            return

        # Handle Escape to close search
        if event.key() == QtCore.Qt.Key_Escape and self.search_widget.isVisible():
            self.search_widget.hide_search()
            return

        # Call parent for other keys
        super(CodeEditor, self).keyPressEvent(event)

    def setup_editor(
        self,
        line_numbers=True,
        markers=True,
        edge_line=100,
        tab_mode=False,
        show_blanks=True,
        font=QtGui.QFontDatabase.systemFont(QtGui.QFontDatabase.FixedFont),
        language="Python",
        filename="",
    ):
        print("setup_editor called")

    def set_color_scheme(self, color_scheme):
        """
        Sets the color theme of the editor widget.
        :param str color_scheme: Name of the color theme to be set
        """

        if color_scheme == "Light":
            self.setStyleSheet("")
            self.setPalette(QtWidgets.QApplication.style().standardPalette())
        else:
            # Now use a palette to switch to dark colors:
            white_color = QColor(255, 255, 255)
            black_color = QColor(0, 0, 0)
            red_color = QColor(255, 0, 0)
            palette = QPalette()
            palette.setColor(QPalette.Window, QColor(53, 53, 53))
            palette.setColor(QPalette.WindowText, white_color)
            palette.setColor(QPalette.Base, QColor(25, 25, 25))
            palette.setColor(QPalette.AlternateBase, QColor(53, 53, 53))
            palette.setColor(QPalette.ToolTipBase, black_color)
            palette.setColor(QPalette.ToolTipText, white_color)
            palette.setColor(QPalette.Text, white_color)
            palette.setColor(QPalette.Button, QColor(53, 53, 53))
            palette.setColor(QPalette.ButtonText, white_color)
            palette.setColor(QPalette.BrightText, red_color)
            palette.setColor(QPalette.Link, QColor(42, 130, 218))
            palette.setColor(QPalette.Highlight, QColor(42, 130, 218))
            palette.setColor(QPalette.HighlightedText, black_color)
            self.setPalette(palette)

    def set_font(self, new_font):
        self.font = new_font
        self.setFont(new_font)

    def toggle_wrap_mode(self, wrap_mode):
        self.setLineWrapMode(wrap_mode)

    def set_cursor_position(self, position):
        """
        Allows the caller to set the position of the cursor within
        the editor text.
        """

        cursor = self.textCursor()
        cursor.setPosition(position)
        self.setTextCursor(cursor)

    def go_to_line(self, line_number):
        """
        Set the text cursor at a specific line number.
        """

        cursor = self.textCursor()

        # Line numbers start at 0
        block = self.document().findBlockByNumber(line_number - 1)

        cursor.setPosition(block.position())
        self.setTextCursor(cursor)

    def toggle_comment(self):
        """
        High level method to comment or uncomment a single line,
        or block of lines.
        """

        # See if there is a selection range
        sel_range = self.get_selection_range()
        if sel_range[0] == 0 and sel_range[1] == 0:
            # Get the text of the line
            cursor = self.textCursor()
            cursor.movePosition(QtGui.QTextCursor.StartOfLine)
            line_text = cursor.block().text()

            # Skip blank lines
            if line_text == "":
                return

            # Find the first non-whitespace character position
            pos = 0
            while pos < len(line_text) and line_text[pos].isspace():
                pos += 1

            # Move right by pos characters to the position before text starts
            cursor.movePosition(
                QtGui.QTextCursor.Right, QtGui.QTextCursor.MoveAnchor, pos
            )
            cursor.movePosition(
                QtGui.QTextCursor.Right, QtGui.QTextCursor.KeepAnchor, 1
            )

            # If we have a single line comment, remove it
            if cursor.selectedText() == "#":
                cursor.removeSelectedText()

                # Remove any whitespace after the comment character
                cursor.movePosition(
                    QtGui.QTextCursor.Right, QtGui.QTextCursor.KeepAnchor, 1
                )
                if cursor.selectedText() == " ":
                    cursor.removeSelectedText()
            else:
                # Insert the comment characters
                cursor.movePosition(
                    QtGui.QTextCursor.Left, QtGui.QTextCursor.MoveAnchor, 1
                )
                cursor.insertText("# ")
        else:
            # Make the selected line numbers 1-based
            sel_start = sel_range[0]
            sel_end = sel_range[1]
            cursor = self.textCursor()

            # Select the text block
            block = self.document().findBlockByNumber(sel_start)
            cursor.setPosition(block.position())
            last_block = self.document().findBlockByNumber(sel_end)
            end_pos = last_block.position() + last_block.length() - 1
            cursor.setPosition(end_pos, QtGui.QTextCursor.KeepAnchor)

            # Find the left-most position to put the comment at
            leftmost_pos = 99999
            comment_line_found = False
            non_comment_line_found = False
            blank_lines = []
            # Step through all of the selected lines and toggle their comments
            for i in range(sel_start, sel_end + 1):
                # Set the cursor to the current line number
                block = self.document().findBlockByNumber(i)
                cursor.setPosition(block.position())
                cursor.movePosition(cursor.EndOfBlock, cursor.KeepAnchor)
                line_text = cursor.selectedText()

                # Make sure the line is not blank
                if line_text == "":
                    blank_lines.append(i)
                    continue

                if line_text.strip()[0] == "#":
                    comment_line_found = True
                else:
                    non_comment_line_found = True

                # Find the first non-whitespace character position
                pos = 0
                while pos < len(line_text) and line_text[pos].isspace():
                    pos += 1

                # Save the left-most position
                if pos < leftmost_pos:
                    leftmost_pos = pos

            # Step through all of the selected lines and toggle their comments
            for i in range(sel_start, sel_end + 1):
                # If this is a blank line, do not process it
                if i in blank_lines:
                    continue

                # Set the cursor to the current line number
                block = self.document().findBlockByNumber(i)
                cursor.setPosition(block.position())

                # See if we need to comment the whole block
                if comment_line_found and non_comment_line_found:
                    # Insert the comment characters
                    cursor.insertText("# ")
                else:
                    # Move right by pos characters to the position before text starts
                    cursor.movePosition(
                        QtGui.QTextCursor.Right,
                        QtGui.QTextCursor.MoveAnchor,
                        leftmost_pos,
                    )
                    cursor.movePosition(
                        QtGui.QTextCursor.Right, QtGui.QTextCursor.KeepAnchor, 1
                    )

                    # If the line starts with a hash, uncomment it. Otherwise comment it
                    if cursor.selectedText() == "#":
                        cursor.removeSelectedText()

                        # Remove any whitespace after the comment character
                        cursor.movePosition(
                            QtGui.QTextCursor.Right, QtGui.QTextCursor.KeepAnchor, 1
                        )
                        if cursor.selectedText() == " ":
                            cursor.removeSelectedText()
                    else:
                        # Insert the comment characters
                        cursor.movePosition(
                            QtGui.QTextCursor.Left, QtGui.QTextCursor.MoveAnchor, 1
                        )
                        cursor.insertText("# ")

    def set_text(self, new_text):
        """
        Sets the text content of the editor.
        :param str new_text: Text to be set in the editor.
        """
        # Set the text in the document
        self.setPlainText(new_text)

        # Set the cursor at the end of the text
        cursor = self.textCursor()
        cursor.movePosition(cursor.End)
        self.setTextCursor(cursor)

        # Set the document to be modified
        self.document().setModified(True)

    def set_text_from_file(self, file_name):
        """
        Allows the editor text to be set from a file.
        :param str file_name: Full path of the file to be loaded into the editor.
        """

        self._filename = file_name

        # Load the text into the text field
        with open(file_name, "r", encoding="utf-8") as file:
            file_content = file.read()

            self.setPlainText(file_content)

    def get_text_with_eol(self):
        """
        Returns a string representing the full text in the editor.
        """
        return self.toPlainText()

    def line_number_area_width(self):
        digits = 1
        max_num = max(1, self.blockCount())
        while max_num >= 10:
            max_num *= 0.1
            digits += 1

        space = 30 + self.fontMetrics().width("9") * digits
        return space

    def resizeEvent(self, e):
        super(CodeEditor, self).resizeEvent(e)
        cr = self.contentsRect()
        width = self.line_number_area_width()
        rect = QtCore.QRect(cr.left(), cr.top(), width, cr.height())
        self.line_number_area.setGeometry(rect)

    def lineNumberAreaPaintEvent(self, event):
        painter = QtGui.QPainter(self.line_number_area)
        try:
            # painter.fillRect(event.rect(), QtCore.Qt.lightGray)
            block = self.firstVisibleBlock()
            block_number = block.blockNumber()
            offset = self.contentOffset()
            top = self.blockBoundingGeometry(block).translated(offset).top()
            bottom = top + self.blockBoundingRect(block).height()

            while block.isValid() and top <= event.rect().bottom():
                if block.isVisible() and bottom >= event.rect().top():
                    line_number = block_number + 1

                    # Draw the breakpoint dot, if there is a breakpoint on this line
                    if self.line_has_breakpoint(line_number):
                        painter.setBrush(
                            QtGui.QBrush(QtGui.QColor(255, 0, 0))
                        )  # Red circle
                        painter.setPen(QtGui.QPen(QtGui.QColor(150, 0, 0)))
                        circle_size = 10
                        circle_x = 5
                        circle_y = (
                            int(top)
                            + (self.fontMetrics().height() - circle_size - 2) // 2
                        )
                        painter.drawEllipse(
                            circle_x, circle_y, circle_size, circle_size
                        )

                    # Draw the line number
                    number = str(line_number)
                    painter.setPen(DARK_BLUE)
                    width = self.line_number_area.width() - 10
                    height = self.fontMetrics().height()
                    painter.drawText(
                        0, int(top), width, height, QtCore.Qt.AlignRight, number
                    )

                block = block.next()
                top = bottom
                bottom = top + self.blockBoundingRect(block).height()
                block_number += 1
        finally:
            painter.end()

    def update_line_number_area_width(self, newBlockCount):
        self.setViewportMargins(self.line_number_area_width(), 0, 0, 0)

    def update_line_number_area(self, rect, dy):
        if dy:
            self.line_number_area.scroll(0, dy)
        else:
            width = self.line_number_area.width()
            self.line_number_area.update(0, rect.y(), width, rect.height())

        if rect.contains(self.viewport().rect()):
            self.update_line_number_area_width(0)

    def highlight_current_line(self):
        extra_selections = list()
        if not self.isReadOnly():
            selection = QtWidgets.QTextEdit.ExtraSelection()
            line_color = DARK_BLUE.lighter(160)
            selection.format.setBackground(line_color)
            selection.format.setProperty(QtGui.QTextFormat.FullWidthSelection, True)

            selection.cursor = self.textCursor()
            selection.cursor.clearSelection()
            extra_selections.append(selection)
        self.setExtraSelections(extra_selections)

    def toggle_breakpoint(self, line_number):
        """
        Toggle breakpoint on/off for a given line number.
        """
        if line_number in self.debugger.breakpoints:
            self.debugger.breakpoints.remove(line_number)
        else:
            self.debugger.breakpoints.append(line_number)

        # Repaint the line number area
        self.line_number_area.update()

    def line_has_breakpoint(self, line_number):
        """
        Checks if a line has a breakpoint.
        """
        return line_number in self.debugger.breakpoints

    def paintEvent(self, event):
        """
        Overrides the default paint event so that we can draw the line length indicator.
        """

        # Call the parent's paintEvent first to render the text
        super(CodeEditor, self).paintEvent(event)

        painter = QtGui.QPainter(self.viewport())
        try:
            # Calculate the x position for the line
            font_metrics = self.fontMetrics()
            char_width = font_metrics.width("M")  # Use 'M' for average character width
            x_position = self.edge_line.columns * char_width + self.contentOffset().x()

            # Only draw if the line is within the visible area
            if 0 <= x_position <= self.viewport().width():
                # Set the pen color (light gray is common)
                painter.setPen(
                    QtGui.QPen(QtGui.QColor(200, 200, 200), 1, QtCore.Qt.SolidLine)
                )

                # Draw the vertical line from top to bottom of the viewport
                painter.drawLine(
                    int(x_position), 0, int(x_position), self.viewport().height()
                )
        finally:
            painter.end()



import logbook as logging
import re

from PyQt5 import QtGui
from PyQt5.QtCore import QObject, pyqtSignal
from PyQt5.QtWidgets import QPlainTextEdit, QAction

from ..mixins import ComponentMixin

from ..icons import icon


def strip_escape_sequences(input_string):
    # Regular expression pattern to match ANSI escape codes
    escape_pattern = re.compile(r"\x1B\[[0-?]*[ -/]*[@-~]")

    # Use re.sub to replace escape codes with an empty string
    clean_string = re.sub(escape_pattern, "", input_string)

    return clean_string


class _QtLogHandlerQObject(QObject):
    sigRecordEmit = pyqtSignal(str)


class QtLogHandler(logging.Handler, logging.StringFormatterHandlerMixin):

    def __init__(self, log_widget, *args, **kwargs):

        super(QtLogHandler, self).__init__(*args, **kwargs)

        log_format_string = (
            "[{record.time:%H:%M:%S%z}] {record.level_name}: {record.message}"
        )

        logging.StringFormatterHandlerMixin.__init__(self, log_format_string)

        self._qobject = _QtLogHandlerQObject()
        self._qobject.sigRecordEmit.connect(log_widget.append)

    def emit(self, record):
        self._qobject.sigRecordEmit.emit(self.format(record) + "\n")


class LogViewer(QPlainTextEdit, ComponentMixin):

    name = "Log viewer"

    def __init__(self, *args, **kwargs):

        super(LogViewer, self).__init__(*args, **kwargs)
        self._MAX_ROWS = 500

        self._actions = {
            "Run": [
                QAction(icon("clear"), "Clear Log", self, triggered=self.clear),
            ]
        }

        self.setReadOnly(True)
        self.setMaximumBlockCount(self._MAX_ROWS)
        self.setLineWrapMode(QPlainTextEdit.NoWrap)

        self.handler = QtLogHandler(self)

    def append(self, msg):
        """Append text to the panel with ANSI escape sequences stipped."""
        self.moveCursor(QtGui.QTextCursor.End)
        self.insertPlainText(strip_escape_sequences(msg))

    def clear_log(self):
        """
        Clear the log content.
        """
        self.clear()



import cadquery as cq
from cadquery.occ_impl.assembly import toCAF

from typing import List, Union
from importlib import reload
from types import SimpleNamespace

from OCP.XCAFPrs import XCAFPrs_AISObject
from OCP.TopoDS import TopoDS_Shape
from OCP.AIS import AIS_InteractiveObject, AIS_Shape
from OCP.Quantity import (
    Quantity_TOC_RGB as TOC_RGB,
    Quantity_Color,
    Quantity_NOC_GOLD as GOLD,
)
from OCP.Graphic3d import Graphic3d_NOM_JADE, Graphic3d_MaterialAspect

from PyQt5.QtGui import QColor

DEFAULT_FACE_COLOR = Quantity_Color(GOLD)
DEFAULT_MATERIAL = Graphic3d_MaterialAspect(Graphic3d_NOM_JADE)


def is_cq_obj(obj):

    from cadquery import Workplane, Shape, Assembly, Sketch

    return isinstance(obj, (Workplane, Shape, Assembly, Sketch))


def find_cq_objects(results: dict):

    return {
        k: SimpleNamespace(shape=v, options={})
        for k, v in results.items()
        if is_cq_obj(v)
    }


def to_compound(
    obj: Union[cq.Workplane, List[cq.Workplane], cq.Shape, List[cq.Shape], cq.Sketch],
):

    vals = []

    if isinstance(obj, cq.Workplane):
        vals.extend(obj.vals())
    elif isinstance(obj, cq.Shape):
        vals.append(obj)
    elif isinstance(obj, list) and isinstance(obj[0], cq.Workplane):
        for o in obj:
            vals.extend(o.vals())
    elif isinstance(obj, list) and isinstance(obj[0], cq.Shape):
        vals.extend(obj)
    elif isinstance(obj, TopoDS_Shape):
        vals.append(cq.Shape.cast(obj))
    elif isinstance(obj, list) and isinstance(obj[0], TopoDS_Shape):
        vals.extend(cq.Shape.cast(o) for o in obj)
    elif isinstance(obj, cq.Sketch):
        if obj._faces:
            vals.append(obj._faces)
        else:
            vals.extend(obj._edges)
    else:
        raise ValueError(f"Invalid type {type(obj)}")

    return cq.Compound.makeCompound(vals)


def to_workplane(obj: cq.Shape):

    rv = cq.Workplane("XY")
    rv.objects = [
        obj,
    ]

    return rv


def make_AIS(
    obj: Union[
        cq.Workplane,
        List[cq.Workplane],
        cq.Shape,
        List[cq.Shape],
        cq.Assembly,
        AIS_InteractiveObject,
    ],
    options={},
):

    shape = None

    if isinstance(obj, cq.Assembly):
        label, shape = toCAF(obj)
        ais = XCAFPrs_AISObject(label)
    elif isinstance(obj, AIS_InteractiveObject):
        ais = obj
    else:
        shape = to_compound(obj)
        ais = AIS_Shape(shape.wrapped)

    set_material(ais, DEFAULT_MATERIAL)
    set_color(ais, DEFAULT_FACE_COLOR)

    if "alpha" in options:
        set_transparency(ais, options["alpha"])
    if "color" in options:
        set_color(ais, to_occ_color(options["color"]))
    if "rgba" in options:
        r, g, b, a = options["rgba"]
        set_color(ais, to_occ_color((r, g, b)))
        set_transparency(ais, a)

    return ais, shape


def export(
    obj: Union[cq.Workplane, List[cq.Workplane]], type: str, file, precision=1e-1
):

    comp = to_compound(obj)

    if type == "stl":
        comp.exportStl(file, tolerance=precision)
    elif type == "step":
        comp.exportStep(file)
    elif type == "brep":
        comp.exportBrep(file)


def to_occ_color(color) -> Quantity_Color:

    if not isinstance(color, QColor):
        if isinstance(color, tuple):
            if isinstance(color[0], int):
                color = QColor(*color)
            elif isinstance(color[0], float):
                color = QColor.fromRgbF(*color)
            else:
                raise ValueError("Unknown color format")
        else:
            color = QColor(color)

    return Quantity_Color(color.redF(), color.greenF(), color.blueF(), TOC_RGB)


def get_occ_color(obj: Union[AIS_InteractiveObject, Quantity_Color]) -> QColor:

    if isinstance(obj, AIS_InteractiveObject):
        color = Quantity_Color()
        obj.Color(color)
    else:
        color = obj

    return QColor.fromRgbF(color.Red(), color.Green(), color.Blue())


def set_color(ais: AIS_Shape, color: Quantity_Color) -> AIS_Shape:

    drawer = ais.Attributes()
    drawer.SetupOwnShadingAspect()
    drawer.ShadingAspect().SetColor(color)

    return ais


def set_material(ais: AIS_Shape, material: Graphic3d_MaterialAspect) -> AIS_Shape:

    drawer = ais.Attributes()
    drawer.SetupOwnShadingAspect()
    drawer.ShadingAspect().SetMaterial(material)

    return ais


def set_transparency(ais: AIS_Shape, alpha: float) -> AIS_Shape:

    drawer = ais.Attributes()
    drawer.SetupOwnShadingAspect()
    drawer.ShadingAspect().SetTransparency(1.0 - alpha)

    return ais


def reload_cq():

    # NB: order of reloads is important
    reload(cq.types)
    reload(cq.occ_impl.geom)
    reload(cq.occ_impl.shapes)
    reload(cq.occ_impl.shapes)
    reload(cq.occ_impl.importers.dxf)
    reload(cq.occ_impl.importers)
    reload(cq.occ_impl.solver)
    reload(cq.occ_impl.assembly)
    reload(cq.occ_impl.sketch_solver)
    reload(cq.hull)
    reload(cq.selectors)
    reload(cq.sketch)
    reload(cq.occ_impl.exporters.svg)
    reload(cq.cq)
    reload(cq.occ_impl.exporters.dxf)
    reload(cq.occ_impl.exporters.amf)
    reload(cq.occ_impl.exporters.json)
    # reload(cq.occ_impl.exporters.assembly)
    reload(cq.occ_impl.exporters)
    reload(cq.assembly)
    reload(cq)


def is_obj_empty(obj: Union[cq.Workplane, cq.Shape]) -> bool:

    rv = False

    if isinstance(obj, cq.Workplane):
        rv = True if isinstance(obj.val(), cq.Vector) else False

    return rv
